import duckdb

from lakefs_spec import LakeFSFileSystem

fs = LakeFSFileSystem()
duckdb.register_filesystem(fs)

with fs.transaction("quickstart", "main") as tx:
    lakes = duckdb.read_parquet("lakefs://quickstart/main/lakes.parquet")
    italian_lakes = duckdb.sql("SELECT * FROM lakes where Country='Italy'")
    italian_lakes.to_csv(f"lakefs://quickstart/{tx.branch.id}/italian_lakes.csv")

    tx.commit(message="Add Italian lakes")



# ---
# jupyter:
#   jupytext:
#     cell_markers: '"""'
#     cell_metadata_filter: -all
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.15.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
"""
# Data Science with lakeFS-spec

In this notebook, we will complete a small end-to-end data science tutorial that employs lakeFS-spec for data versioning.
We will use versioned weather data to train a decision tree classifier to predict whether it is raining tomorrow given the current weather.

We will do the following:

* Environment setup
* LakeFS setup
* Authenticating with the lakeFS server
* Data ingestion via transactions
* Model training
* Updating data and retraining a model
* Accessing data versions and reproducing experiments
* Using tags for semantic versioning

!!! tip "Local Execution"
    If you want to execute the code in this tutorial as a Jupyter notebook yourself, download the `demo_data_science_project.py` file from the lakeFS-spec repository.

    You can then convert the Python file to a notebook using [Jupytext](https://jupytext.readthedocs.io/en/latest/using-cli.html) using the following command: `jupytext --to notebook demo_data_science_project.py`.

This tutorial assumes that you have installed lakeFS-spec in a virtual environment, and that you have followed the [quickstart guide](../quickstart.md) to set up a local lakeFS instance.

## Environment setup

Install the necessary libraries for this notebook on the environment you have just created:
"""

# %% tags=["Remove_single_output"]
# %pip install numpy pandas scikit-learn

# %% [markdown]
"""
Also install an appropriate lakeFS-spec version, which can be either the latest release from PyPI via `pip install --upgrade lakefs-spec`, or the development version from GitHub via `pip install git+https://github.com/aai-institute/lakefs-spec.git`.
"""

# %% [markdown]
"""
## lakeFS Setup

With Docker Desktop or a similar runtime running set up lakeFS by executing the following `docker run` command (from the [lakeFS quickstart](https://docs.lakefs.io/quickstart/launch.html)) in your console:

```shell
docker run --name lakefs --pull always --rm --publish 8000:8000 treeverse/lakefs:latest run --quickstart
```

You find the authentication credentials in the terminal output. The default address for the local lakeFS GUI is http://localhost:8000/.

## Authenticating with the lakeFS server

There are multiple ways to authenticate with lakeFS from Python code. In this tutorial, we choose the YAML file configuration.
By executing the cell below, you will download a YAML file containing the default lakeFS quickstart credentials and server URL to your user directory.
"""

# %% tags=["Remove_all_output"]
import os
import tempfile
import urllib.request
from pathlib import Path

urllib.request.urlretrieve(
    "https://raw.githubusercontent.com/aai-institute/lakefs-spec/main/docs/tutorials/.lakectl.yaml",
    os.path.expanduser("~/.lakectl.yaml"),
)

# %% [markdown]
"""
We can now instantiate the `LakeFSFileSystem` with the credentials we just downloaded.
Alternatively, we could have passed the credentials directly in the code.
It is important that the credentials are available at the time of filesystem instantiation.
"""

# %%
from lakefs_spec import LakeFSFileSystem

fs = LakeFSFileSystem()

REPO_NAME = "weather"

# %% [markdown]
"""
We will create a repository using a helper function provided by lakeFS-spec.
If you have already created one in the UI, make sure to set the `REPO_NAME` variable accordingly in the cell directly above.
"""

# %%
import lakefs

repo = lakefs.Repository(REPO_NAME, fs.client).create(storage_namespace=f"local://{REPO_NAME}")

# %% [markdown]
"""
## Data Ingestion

Now it's time to get some data. We will use the [Open-Meteo API](https://open-meteo.com/), where we can pull weather data from an API for free (as long as we are non-commercial) and without an API token.
In order to prevent hitting the rate limits when repeatedly querying the API (and out of courtesy towards the operators of the API), the `_maybe_urlretrieve` function provides a simple local cache for the downloaded data.

For training our toy model, we download the full weather data of Munich for the year 2010:
"""


# %%
def _maybe_urlretrieve(url: str, filename: str) -> str:
    # Avoid API rate limit errors by downloading to a fixed local location
    destination = Path(tempfile.gettempdir()) / "lakefs-spec-tutorials" / filename
    destination.parent.mkdir(exist_ok=True, parents=True)
    if destination.exists():
        return str(destination)

    outfile, _ = urllib.request.urlretrieve(url, str(destination))
    return outfile


outfile = _maybe_urlretrieve(
    "https://archive-api.open-meteo.com/v1/archive?latitude=52.52&longitude=13.41&start_date=2010-01-01&end_date=2010-12-31&hourly=temperature_2m,relativehumidity_2m,rain,pressure_msl,surface_pressure,cloudcover,cloudcover_low,cloudcover_mid,cloudcover_high,windspeed_10m,windspeed_100m,winddirection_10m,winddirection_100m",
    "weather-2010.json",
)

# %% [markdown]
"""
The data is in JSON format.
Therefore, we need to wrangle the data a bit to make it usable.
But first, we will upload it to our lakeFS instance.
"""

# %% [markdown]
"""
## Upload a file using transactions

lakeFS works similar to `git` as a versioning system.
You can create *commits* that contain specific changes to the data.
You can also work with *branches* to create your own isolated view of the data independently of your colleagues.
Every commit (on any branch) is identified by a commit SHA.
This SHA can be used to programmatically interact with specific states of your data and enables logging of the specific data versions used to create a certain model.

To easily carry out versioning operations while uploading files, you can use **transactions**.
A transaction is a context manager that keeps track of all files that were uploaded in its scope, as well as all versioning operations happening between file uploads.
All operations are deferred to the end of the transaction, and are executed sequentially on completion.

To create a commit after a file upload, you can run the following transaction:
"""

# %%
NEW_BRANCH = lakefs.Branch(REPO_NAME, "transform-raw-data", client=fs.client)
NEW_BRANCH.create("main")

with fs.transaction(REPO_NAME, NEW_BRANCH) as tx:
    fs.put(outfile, f"{REPO_NAME}/{tx.branch.id}/weather-2010.json")
    tx.commit(message="Add 2010 weather data")

# %% [markdown]
"""
You can inspect this commit by selecting the `transform-raw-data` branch, and navigating to the **Commits** tab.
"""

# %% [markdown]
"""
## Data Transformation

Now let's transform the data for our use case.
We put the transformation into a function to be able to reuse it later.

In this notebook, we use a simple toy model to predict whether it is raining at the same time tomorrow given weather data from right now.

We will skip a lot of possible feature engineering and other data science aspects in order to focus more on the application of the `LakeFSFileSystem`.
"""

# %%
import json

import pandas as pd


def transform_json_weather_data(filepath):
    if hasattr(filepath, "close") and hasattr(filepath, "tell"):
        data = json.load(filepath)
    else:
        with open(filepath, "r") as f:
            data = json.load(f)

    df = pd.DataFrame.from_dict(data["hourly"])
    df.time = pd.to_datetime(df.time)
    df["is_raining"] = df.rain > 0
    df["is_raining_in_1_day"] = df.is_raining.shift(24).astype(bool)
    df = df.dropna()
    return df


df = transform_json_weather_data(outfile)
df.head(5)

# %% [markdown]
"""
Next, we save this data as a CSV file into the main branch.
When the transaction commit helper is called, the newly put CSV file is committed.
You can verify the saving worked in the lakeFS UI in your browser by switching to the commits tab of the `main` branch.
"""

# %%
with fs.transaction(REPO_NAME, "main") as tx:
    df.to_csv(f"lakefs://{REPO_NAME}/{tx.branch.id}/weather_2010.csv")
    tx.commit(message="Update weather data")

# %% [markdown]
"""
## Training the initial weather model

First we will do a train-test split:
"""

# %%
import sklearn.model_selection

model_data = df.drop("time", axis=1)

train, test = sklearn.model_selection.train_test_split(model_data, random_state=7)

# %% [markdown]
"""
We save these train and test datasets into a new `training` branch.
If the branch does not exist yet, as in this case, it is implicitly created by default.
You can control this behaviour with the `create_branch_ok` flag when initializing the `LakeFSFileSystem`.
By default, `create_branch_ok` is set to `True`, so we need to only set `fs = LakeFSFileSystem()` to enable implicit branch creation.
"""

# %%
TRAINING_BRANCH = lakefs.Branch(REPO_NAME, "training", client=fs.client)
TRAINING_BRANCH.create("main")

with fs.transaction(REPO_NAME, TRAINING_BRANCH) as tx:
    train.to_csv(f"lakefs://{REPO_NAME}/{tx.branch.id}/train_weather.csv")
    test.to_csv(f"lakefs://{REPO_NAME}/{tx.branch.id}/test_weather.csv")
    tx.commit(message="Add train-test split of 2010 weather data")

# %% [markdown]
"""
Let's check the shape of train and test data.
Later on, we will get back to this data version and reproduce the results of the experiment.
"""

# %%
print(f"Initial train data shape: {train.shape}")
print(f"Initial test data shape: {test.shape}")

# %% [markdown]
"""
We now proceed to train a decision tree classifier and evaluate it on the test set:
"""

# %%
from sklearn.tree import DecisionTreeClassifier

dependent_variable = "is_raining_in_1_day"

model = DecisionTreeClassifier(random_state=7)

x_train, y_train = train.drop(dependent_variable, axis=1), train[dependent_variable].astype(bool)
x_test, y_test = test.drop(dependent_variable, axis=1), test[dependent_variable].astype(bool)

model.fit(x_train, y_train)

test_acc = model.score(x_test, y_test)
print(f"Test accuracy: {test_acc:.2%}")

# %% [markdown]
"""
## Updating data and retraining the model

Until now, we only have used data from 2010.
Let's download additional 2020 data, transform it, and save it to lakeFS.
"""

# %%
outfile = _maybe_urlretrieve(
    "https://archive-api.open-meteo.com/v1/archive?latitude=52.52&longitude=13.41&start_date=2020-01-01&end_date=2020-12-31&hourly=temperature_2m,relativehumidity_2m,rain,pressure_msl,surface_pressure,cloudcover,cloudcover_low,cloudcover_mid,cloudcover_high,windspeed_10m,windspeed_100m,winddirection_10m,winddirection_100m",
    "weather-2020.json",
)

new_data = transform_json_weather_data(outfile)

with fs.transaction(REPO_NAME, "main") as tx:
    new_data.to_csv(f"lakefs://{REPO_NAME}/{tx.branch.id}/weather_2020.csv")
    tx.commit(message="Add 2020 weather data")

# Remove leftover temporary files from previous `urlretrieve` calls
urllib.request.urlcleanup()


# %% [markdown]
"""
Let's concatenate the old data and the new data, create a new train-test split, and push the updated files to lakeFS:
"""

# %%
new_data = new_data.drop("time", axis=1)
full_data = pd.concat([new_data, train, test])

train_df, test_df = sklearn.model_selection.train_test_split(full_data, random_state=7)

print(f"Updated train data shape: {train_df.shape}")
print(f"Updated test data shape: {test_df.shape}")

with fs.transaction(REPO_NAME, TRAINING_BRANCH) as tx:
    train_df.to_csv(f"lakefs://{REPO_NAME}/{tx.branch.id}/train_weather.csv")
    test_df.to_csv(f"lakefs://{REPO_NAME}/{tx.branch.id}/test_weather.csv")
    tx.commit(message="Add train-test split of 2010 and 2020 data")

# %% [markdown]
"""
Now, we train the model on the new data and validate on the new test data.
"""

# %%
x_train, y_train = (
    train_df.drop(dependent_variable, axis=1),
    train_df[dependent_variable].astype(bool),
)
x_test, y_test = test_df.drop(dependent_variable, axis=1), test_df[dependent_variable].astype(bool)

model.fit(x_train, y_train)

test_acc = model.score(x_test, y_test)

print(f"Test accuracy: {test_acc:.2%}")

# %% [markdown]
"""
## Accessing data versions through commits and reproducing experiments

If we need to go to our initial data and reproduce the first experiment (the model trained on the 2010 data with its initial accuracy), we can go back in the commit history of the `training` branch and select the appropriate commit data snapshot.
Since we have created multiple commits on the same branch already, we will address different data versions by their commit SHAs.

To obtain the actual commit SHA from a branch, we have multiple options.
Manually, we could go into the lakeFS UI, select the training branch, and navigate to the **Commits** tab.
There, we take the parent of the previous commit, titled `Add train-test split of 2010 weather data`, and copy its revision SHA (also called `ID`).

In code, we can obtain commit SHAs for different revisions on the `training` branch by using `lakefs.Reference` objects.
"""

# %%

# access the data of the previous commit with a lakefs ref expression, in this case the same as in git.
previous_commit = repo.ref(f"{TRAINING_BRANCH.id}~").get_commit()
fixed_commit_id = previous_commit.id
print(fixed_commit_id)

# %% [markdown]
"""
Let's check whether we managed to get the initial train and test data with this commit SHA, checking equality to the initial data:
"""

# %%
orig_train = pd.read_csv(f"lakefs://{REPO_NAME}/{fixed_commit_id}/train_weather.csv", index_col=0)
orig_test = pd.read_csv(f"lakefs://{REPO_NAME}/{fixed_commit_id}/test_weather.csv", index_col=0)

print(f"Is the pulled training data equal to the local training data? {train.equals(orig_train)}")
print(f"Is the pulled test data equal to the local test data? {test.equals(orig_test)}")

# %% [markdown]
"""
Let's train and validate the model again based on the redownloaded data and see if we manage to reproduce the initial accuracy.
"""
# %%
x_train, y_train = train.drop(dependent_variable, axis=1), train[dependent_variable].astype(bool)
x_test, y_test = test.drop(dependent_variable, axis=1), test[dependent_variable].astype(bool)

model.fit(x_train, y_train)

test_acc = model.score(x_test, y_test)

print(f"Test accuracy: {test_acc:.2%}")

# %% [markdown]
"""
## Using tags instead of commit SHAs for semantic versioning

The above method for data versioning works great when you have experiment tracking tools to store and retrieve the commit SHA in automated pipelines.
But it can be tedious to retrieve in manual prototyping.
We can make selected versions of the dataset more accessible with semantic versioning by attaching a human-interpretable tag to a specific commit SHA.

Creating a tag is easiest when done inside a transaction, just like the files we already uploaded.
To do this, simply call `tx.tag` on the transaction and supply the repository name, the commit SHA to tag, and the intended tag name.
Tags are immutable once created, so attempting to tag two different commits with the same name will result in an error.
"""


# %%
with fs.transaction(REPO_NAME, "main") as tx:
    # returns the tag as a lakeFS object.
    tag = tx.tag(fixed_commit_id, name="train-test-split-2010")


# %% [markdown]
"""
Now we can access the specific files with the semantic tag.
Both the `fixed_commit_id` and `tag` reference the same version `ref` in lakeFS, whereas a branch name always points to the latest version on that respective branch.
"""

# %%
train_from_commit = pd.read_csv(
    f"lakefs://{REPO_NAME}/{fixed_commit_id}/train_weather.csv", index_col=0
)
train_from_tag = pd.read_csv(f"lakefs://{REPO_NAME}/{tag.id}/train_weather.csv", index_col=0)

# %% [markdown]
"""
We can verify this by comparing the `DataFrame`s. We see that the `train_from_commit` and `train_from_tag` are equal.
"""

# %%
print(
    f"Is the data tagged {tag!r} equal to the data in commit {fixed_commit_id[:8]}? {train_from_commit.equals(train_from_tag)}"
)

# %% tags=["Remove_input", "Remove_all_output"]
# Clean-up cell removing artifacts created in notebook execution to ensure idempotency.
# Cell hidden in docs

for t in repo.tags():
    t.delete()
for b in repo.branches():
    if b.id != "main":
        b.delete()



import random
import string

import pytest
from lakefs.branch import Branch
from lakefs.repository import Repository

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory, list_branchnames, put_random_file_on_branch, with_counter


def test_no_change_postcommit(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    message = f"Add file {random_file.name}"

    with fs.transaction(repository, temp_branch) as tx:
        fs.put(lpath, f"{repository.id}/{tx.branch.id}/{random_file.name}")
        tx.commit(message=message)

    commits = list(temp_branch.log(max_amount=2))
    current_head = temp_branch.head.get_commit()
    assert commits[0].message.startswith("Merge")
    assert commits[1].message == message

    # put the same file again, this time the diff is empty
    with fs.transaction(repository, temp_branch) as tx:
        fs.put(lpath, f"{repository.id}/{tx.branch.id}/{random_file.name}", precheck=False)
        tx.commit(message=f"Add file {random_file.name}")

    # check that no other commit has happened.
    assert temp_branch.head.get_commit() == current_head


def test_implicit_branch_creation(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    non_existing_branch = "non-existing-" + "".join(random.choices(string.digits, k=8))
    put_random_file_on_branch(random_file_factory, fs, repository, non_existing_branch)

    assert non_existing_branch in list_branchnames(repository)
    # branch has been created at this point
    repository.branch(non_existing_branch).delete()

    fs.create_branch_ok = False
    another_non_existing_branch = "non-existing-" + "".join(random.choices(string.digits, k=8))
    with pytest.raises(FileNotFoundError, match="Not Found: .*"):
        put_random_file_on_branch(random_file_factory, fs, repository, another_non_existing_branch)


def test_put_client_caching(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    """
    Tests that ``precheck=True`` prevents a second upload of an identical file by matching checksums.
    """
    fs.client, counter = with_counter(fs.client)

    rpath = put_random_file_on_branch(random_file_factory, fs, repository, temp_branch)
    assert fs.exists(rpath)



import lakefs
import pytest
from lakefs.repository import Repository

from lakefs_spec import LakeFSFileSystem


def test_info_on_directory(fs: LakeFSFileSystem, repository: Repository) -> None:
    """
    Check that an `fs.info` call on a trailing-slash resource yields a directory.
    """
    resource = f"{repository.id}/main/"
    res = fs.info(resource)

    assert res["type"] == "directory"


def test_info_on_nonexistent_directory(fs: LakeFSFileSystem, repository: Repository) -> None:
    """
    Check that a nonexistent directory raises a FileNotFoundError in `fs.info`.
    """
    resource = f"{repository.id}/main/blabla/"

    with pytest.raises(FileNotFoundError):
        fs.info(resource)


def test_info_on_directory_no_trailing_slash(fs: LakeFSFileSystem, repository: Repository) -> None:
    """
    Regression test to check that calling `fs.info()` on existing non-slash-terminated
    directories yields the same results as if terminated with a slash.
    """
    resource = f"{repository.id}/main/data/"
    res = fs.info(resource)

    assert res["type"] == "directory"

    non_slash_resource = resource.rstrip("/")
    assert fs.info(non_slash_resource) == res


def test_info_on_commit(
    fs: LakeFSFileSystem,
    repository: Repository,
) -> None:
    prefix = f"lakefs://{repository.id}"

    head = lakefs.Branch(repository.id, "main", client=fs.client).head

    binfo = fs.info(f"{prefix}/main/README.md")
    branch_metadata = (binfo["checksum"], binfo["mtime"], binfo["size"])
    # fetching directly from commit should yield the same result.
    cinfo = fs.info(f"{prefix}/{head.id}/README.md")
    commit_metadata = (cinfo["checksum"], cinfo["mtime"], cinfo["size"])

    assert branch_metadata == commit_metadata



from __future__ import annotations

import collections
import random
import string
import uuid
from inspect import ismethod
from pathlib import Path

from lakefs.branch import Branch
from lakefs.client import Client
from lakefs.repository import Repository

from lakefs_spec import LakeFSFileSystem


class APICounter:
    def __init__(self):
        self._counts: dict[str, int] = collections.defaultdict(int)

    def clear(self):
        self._counts.clear()

    def count(self, name: str) -> int:
        return self._counts[name]

    def counts(self):
        """Gives an iterator over the API counts."""
        return self._counts.values()

    def named_counts(self):
        """Gives an iterator over the API names and counts."""
        return self._counts.items()

    def increment(self, name: str) -> None:
        self._counts[name] += 1


def with_counter(client: Client) -> tuple[Client, APICounter]:
    """Instruments a lakeFS API client with an API counter."""
    counter = APICounter()

    def patch(fn, name):
        """Patches an API instance method ``fn`` on an API ``apiname``."""

        def wrapped_fn(*args, **kwargs):
            counter.increment(name)
            return fn(*args, **kwargs)

        return wrapped_fn

    for api_name, api in client.sdk_client.__dict__.items():
        if api_name == "_api":
            continue

        for ep_name in filter(
            lambda op: not op.startswith("_")
            and not op.endswith("with_http_info")
            and ismethod(getattr(api, op)),
            dir(api),
        ):
            endpoint = getattr(api, ep_name)
            setattr(api, ep_name, patch(endpoint, f"{api_name}.{ep_name}"))

    return client, counter


class RandomFileFactory:
    def __init__(self, path: str | Path):
        path = Path(path)
        if not path.is_dir():
            raise NotADirectoryError(path)
        self.path = path

    def list(self) -> list[Path]:
        return list(self.path.iterdir())

    def make(self, fname: str | None = None, size: int = 2**10) -> Path:
        """
        Generate a random file named ``fname`` with a random string of size
        ``size`` (in bytes) as content.
        """
        if fname is None:
            fname = "test-" + str(uuid.uuid4()) + ".txt"
        random_file = self.path / fname
        random_str = "".join(random.choices(string.ascii_letters + string.digits, k=size))
        random_file.write_text(random_str, encoding="utf-8")
        return random_file


def put_random_file_on_branch(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: str | Branch,
    commit: bool = False,
) -> str:
    random_file = random_file_factory.make()
    branch = temp_branch.id if isinstance(temp_branch, Branch) else temp_branch
    lpath = str(random_file)
    rpath = f"{repository.id}/{branch}/{random_file.name}"

    fs.put(lpath, rpath, precheck=False)
    if commit:
        if isinstance(temp_branch, str):
            temp_branch = repository.branch(temp_branch)
        temp_branch.commit(message=f"Add file {random_file.name!r}")
    return rpath


def list_branchnames(repository: Repository) -> list[str]:
    return [branch.id for branch in repository.branches()]



import filecmp
from pathlib import Path

import pytest
from lakefs.branch import Branch
from lakefs.repository import Repository

from lakefs_spec.spec import LakeFSFileSystem
from tests.util import RandomFileFactory


def test_walk_single_dir(fs: LakeFSFileSystem, repository: Repository) -> None:
    """`walk` in a single directory should find all files contained therein"""
    branch = "main"
    resource = "images"
    path = f"{repository.id}/{branch}/{resource}/"

    dirname, dirs, files = next(fs.walk(path))
    assert dirname == path
    assert dirs == []
    assert len(files) == 37  # NOTE: hardcoded for quickstart repo


def test_walk_repo_root(fs: LakeFSFileSystem, repository: Repository) -> None:
    """`walk` should be able to be called on the root directory of a repository"""
    branch = "main"
    path = f"{repository.id}/{branch}/"

    dirname, dirs, files = next(fs.walk(path))
    assert dirname == path
    assert len(dirs) == 2
    assert len(files) == 2


def test_find_in_folder(fs: LakeFSFileSystem, repository: Repository) -> None:
    path = f"{repository.id}/main/"
    # Find the 37 elements in images directory in test repo
    assert len(fs.find(path + "images")) == 37


def test_touch(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    rpath = f"{repository.id}/{temp_branch.id}/hello.txt"
    fs.touch(rpath)
    assert fs.exists(rpath)

    # mock the server config to an older version.
    fs._lakefs_server_version = (1, 0, 0)

    with pytest.raises(NotImplementedError, match=r"\.touch\(\) is not supported"):
        fs.touch(rpath)


def test_glob(
    fs: LakeFSFileSystem,
    repository: Repository,
) -> None:
    branch = "main"
    files = fs.glob(f"lakefs://{repository.id}/{branch}/**/*.png")
    assert len(files) > 0


def test_du(
    fs: LakeFSFileSystem,
    repository: Repository,
) -> None:
    branch = "main"
    size = fs.du(f"lakefs://{repository.id}/{branch}/", withdirs=True)
    assert size > 2**20  # at least 1 MiB in the quickstart repo


def test_size(
    fs: LakeFSFileSystem,
    repository: Repository,
) -> None:
    branch = "main"
    size = fs.size(f"lakefs://{repository.id}/{branch}/lakes.parquet")
    assert size >= 2**19  # lakes.parquet is larger than 500 KiB


def test_isfile_isdir(
    fs: LakeFSFileSystem,
    repository: Repository,
) -> None:
    branch = "main"
    assert fs.isfile(f"lakefs://{repository.id}/{branch}/lakes.parquet")
    assert not fs.isdir(f"lakefs://{repository.id}/{branch}/lakes.parquet")

    assert not fs.isfile(f"lakefs://{repository.id}/{branch}/data")
    assert fs.isdir(f"lakefs://{repository.id}/{branch}/data")


def test_write_text_read_text(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    rpath = f"lakefs://{repository.id}/{temp_branch.id}/new-file.txt"
    content = "Hello, World!"
    encoding = "utf-32-le"  # use a non-standard encoding

    fs.write_text(rpath, content, encoding=encoding)
    actual = fs.read_text(rpath, encoding=encoding)
    assert actual == content


def test_cat_pipe(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    rpath = f"lakefs://{repository.id}/{temp_branch.id}/new-file.txt"
    content = "Hello, World!"
    encoding = "utf-32-le"  # use a non-standard encoding
    fs.pipe(rpath, content.encode(encoding))

    actual = fs.cat(rpath)
    assert str(actual, encoding=encoding) == content

    actual = fs.cat_file(rpath, end=4)  # Only fetch first UTF-32 glyph
    assert str(actual, encoding=encoding) == content[0]


def test_cat_pipe_file(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    rpath = f"lakefs://{repository.id}/{temp_branch.id}/new-file.txt"
    content = "Hello, World!"
    encoding = "utf-32-le"  # use a non-standard encoding
    fs.pipe_file(rpath, content.encode(encoding))

    actual = fs.cat(rpath)
    assert str(actual, encoding=encoding) == content


def test_cat_ranges(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    rpaths = [f"lakefs://{repository.id}/{temp_branch.id}/file-{idx}.txt" for idx in range(2)]
    content = "Hello, World!"
    encoding = "utf8"

    fs.write_text(rpaths[0], content, encoding=encoding)
    fs.write_text(rpaths[1], content, encoding=encoding)

    # fetch first byte of each file
    ranges = fs.cat_ranges(rpaths, starts=0, ends=1)
    for idx in range(2):
        assert str(ranges[idx], encoding=encoding) == content[0]


def test_head_tail(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    rpath = f"lakefs://{repository.id}/{temp_branch.id}/new-file.txt"
    content = "Hello, World!"
    encoding = "utf-8"

    fs.write_text(rpath, content, encoding=encoding)

    size = 2
    head = fs.head(rpath, size=size)
    assert str(head, encoding=encoding) == content[:size]

    tail = fs.tail(rpath, size=size)
    assert str(tail, encoding=encoding) == content[len(content) - size :]


def test_mv(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    random_file = random_file_factory.make()
    lpath = str(random_file)
    rpath1 = f"{repository.id}/{temp_branch.id}/new_dir/{random_file.name}"
    rpath2 = f"{repository.id}/{temp_branch.id}/{random_file.name}"

    fs.put_file(lpath=lpath, rpath=rpath1)
    assert fs.exists(rpath1)
    assert not fs.exists(rpath2)

    fs.mv(rpath1, rpath2)
    assert not fs.exists(rpath1)
    assert fs.exists(rpath2)


def test_copy(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    random_file = random_file_factory.make()
    lpath = str(random_file)
    rpath1 = f"{repository.id}/{temp_branch.id}/new_dir/{random_file.name}"
    rpath2 = f"{repository.id}/{temp_branch.id}/{random_file.name}"

    fs.put_file(lpath=lpath, rpath=rpath1)
    assert fs.exists(rpath1)
    assert not fs.exists(rpath2)

    fs.cp(rpath1, rpath2)
    assert fs.exists(rpath1)
    assert fs.exists(rpath2)

    assert fs.info(rpath1)["checksum"] == fs.info(rpath2)["checksum"]


def test_get_file(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
    tmp_path: Path,
) -> None:
    random_file = random_file_factory.make()
    lpath1 = str(random_file)
    rpath = f"{repository.id}/{temp_branch.id}/{random_file.name}"
    fs.put(lpath=lpath1, rpath=rpath)

    lpath2 = str(tmp_path / random_file.name)
    fs.get(rpath=rpath, lpath=lpath2)
    assert filecmp.cmp(lpath1, lpath2)



import pandas as pd

from lakefs_spec import LakeFSFileSystem

fs = LakeFSFileSystem()

with fs.transaction("quickstart", "main") as tx:
    lakes = pd.read_parquet(f"lakefs://quickstart/{tx.branch.id}/lakes.parquet")
    german_lakes = lakes.query('Country == "Germany"')
    german_lakes.to_csv(f"lakefs://quickstart/{tx.branch.id}/german_lakes.csv")

    tx.commit(message="Add German lakes")



"""
Error translation facilities to map lakeFS API errors to Python-native OS errors in the lakeFS file system.

This is important to honor the fsspec API contract, where users only need to expect builtin Python exceptions to
avoid complicated error handling setups.
"""

from __future__ import annotations

import errno
from functools import partial

from lakefs.exceptions import ServerException

HTTP_CODE_TO_ERROR: dict[int, type[OSError] | partial[OSError]] = {
    400: partial(IOError, errno.EINVAL),
    401: PermissionError,
    403: PermissionError,
    404: FileNotFoundError,
    410: FileNotFoundError,  # Gone (temporarily / permanently unavailable)
    416: partial(IOError, errno.EINVAL),  # invalid range
    420: partial(IOError, errno.EBUSY),  # too many requests
}


def translate_lakefs_error(
    error: ServerException,
    rpath: str | None = None,
    message: str | None = None,
    set_cause: bool = True,
) -> OSError:
    """
    Convert a lakeFS server exception to a Python builtin exception.

    For some subclasses of ``lakefs.exceptions.ServerException``, a direct Python builtin equivalent exists.
    In these cases, the suitable equivalent is returned. All other classes are converted to a standard ``IOError``.

    Parameters
    ----------
    error: ServerException
        The exception returned by the lakeFS SDK wrapper.
    rpath: str | None
        The remote resource path involved in the error.
    message: str | None
        An error message to use for the returned exception.
         If not given, the error message returned by the lakeFS server is used instead.
    set_cause: bool
        Whether to set the ``__cause__`` attribute to the previous exception if the exception is translated.

    Returns
    -------
    OSError
        A builtin Python exception ready to be thrown.
    """
    status = error.status_code

    if hasattr(error, "body"):
        # error has a JSON response body attached
        reason = error.body["message"]
    else:
        reason = error.reason

    emsg = f"{status} {reason}"
    if rpath:
        emsg += f": {rpath!r}"

    constructor = HTTP_CODE_TO_ERROR.get(status, partial(IOError, errno.EIO))
    custom_exc = constructor(message or emsg)

    if set_cause:
        custom_exc.__cause__ = error
    return custom_exc
